;;; pattern.scm
;;; A [very] simple (but useful) pattern matcher in Scheme
;;;
;;; Programmer: Mayer Goldberg, 2002
;;;
;;; Packaged as a Chicken extension by Category 5, May 2004

; ----------------------------TERMS OF USE--------------------------------
; Copyright (c) 2002-2004, Mayer Goldberg 
; All rights reserved.
;
; Redistribution and use in source and binary forms, with or without
; modification, are permitted provided that the following conditions are
; met:
;
;   Redistributions of source code must retain the above copyright notice,
;   this list of conditions and the following disclaimer. Redistributions in
;   binary form must reproduce the above copyright notice, this list of
;   conditions and the following disclaimer in the documentation and/or
;   other materials provided with the distribution. Neither the name of the
;   author nor the names of its contributors may be used to endorse or
;   promote products derived from this software without specific prior
;   written permission.
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
; IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
; THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
; PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
; CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
; EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
; PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
; LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
; NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
; SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
; ----------------------------TERMS OF USE--------------------------------


(declare
 (usual-integrations)
 (fixnum-arithmetic)
 (no-bound-checks)
 (export pmatch map-match ? with match-struct compose-match compose-match-list)
 (bound-to-procedure pmatch map-match ? with match-struct compose-match compose-match-list))

(register-feature! 'pmatch)

(define pmatch
  (letrec ((pmatch
	    (lambda (s pat ret-match ret-fail)
	      (cond ((pair? pat)
		     (if (pair? s)
			 (pmatch (car s) (car pat)
			  (lambda (m-car)
			    (pmatch (cdr s) (cdr pat)
			     (lambda (m-cdr)
			       (ret-match (append m-car m-cdr)))
			     ret-fail))
			  ret-fail)
			 (ret-fail)))
		    ((vector? pat)
                     (if (vector? s)
                         (and (= (vector-length pat) (vector-length s))
                              (map-match (vector->list s) (vector->list pat)
                                         ret-match ret-fail))
                         (ret-fail)))
		    ((procedure? pat)
		     (if (pat s) (ret-match (list s)) (ret-fail)))
		    ((equal? pat s) (ret-match '()))
		    (else (ret-fail))))))
    (lambda (s pat ret-match ret-fail)
      (pmatch s pat
       (lambda (s) (with s ret-match))
       (lambda () (ret-fail s))))))

(define map-match
  (lambda (arglist patlist ret-match ret-fail)
    (if (null? patlist) (ret-match '()) ; no need to test (null? arglist) !!!
	(pmatch (car arglist) (car patlist)
	 (lambda (m-car)
	   (map-match (cdr arglist) (cdr patlist)
	    (lambda (m-cdr) (ret-match (append m-car m-cdr)))
	    ret-fail))
	 ret-fail))))

(define ?
  (lambda (id . list-pred)
    (let ((ok?
	   (if (null? list-pred)
	       (lambda (x) #t)
	       (car list-pred))))
      ok?)))

(define with (lambda (s f) (apply f s)))

(define match-struct
  (lambda (pattern ret-matched)
    (lambda (e ret-unmatched)
      (pmatch e pattern ret-matched ret-unmatched))))

(define compose-match (lambda args (compose-match-list args)))

(define compose-match-list
  (lambda (s)
    (if (null? s)
	(lambda (e ret-unmatched)
	  (ret-unmatched e ret-unmatched))
	(let ((f (car s))
	      (g (compose-match-list (cdr s))))
	  (lambda (e ret-unmatched)
	    (f e (lambda (e) (g e ret-unmatched))))))))

